@@ -3,7 +3,7 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2015 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2016 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -108,24 +108,12 @@ extern int errno;
 #undef CHARBITS
 #undef INTBITS
 
-#if !defined(ZOS_USS)
 #if HAVE_INTTYPES_H
 # include <inttypes.h>
 #endif
 #if HAVE_STDINT_H
 # include <stdint.h>
 #endif
-#else /* ZOS_USS */
-#include <limits.h>
-#include <sys/time.h>
-#define INT32_MAX INT_MAX
-#define INT32_MIN INT_MIN
-#ifndef __uint32_t
-#define __uint32_t 1
-typedef  unsigned long uint32_t;
-#endif
-typedef  long int32_t;
-#endif /* !ZOS_USS */
 
 /* ----------------- System dependencies (with more includes) -----------*/
 
@@ -159,11 +147,10 @@ typedef int off_t;
 #ifdef NEED_MEMORY_H
 #include <memory.h>
 #endif	/* NEED_MEMORY_H */
-#else	/* not HAVE_STRING_H */
+#endif /* HAVE_STRING_H */
 #ifdef HAVE_STRINGS_H
 #include <strings.h>
 #endif	/* HAVE_STRINGS_H */
-#endif	/* not HAVE_STRING_H */
 
 #if HAVE_UNISTD_H
 #include <unistd.h>
@@ -404,6 +391,45 @@ typedef struct exp_node {
 #		define	MALLOC	0x0001       /* can be free'd */
 
 /* type = Node_val */
+	/*
+	 * STRING and NUMBER are mutually exclusive. They represent the
+	 * type of a value as assigned.
+	 *
+	 * STRCUR and NUMCUR are not mutually exclusive. They represent that
+	 * the particular type of value is up to date.  For example,
+	 *
+	 * 	a = 5		# NUMBER | NUMCUR
+	 * 	b = a ""	# Adds STRCUR to a, since a string value
+	 * 			# is now available. But the type hasn't changed!
+	 *
+	 * 	a = "42"	# STRING | STRCUR
+	 * 	b = a + 0	# Adds NUMCUR to a, since numeric value
+	 * 			# is now available. But the type hasn't changed!
+	 *
+	 * MAYBE_NUM is the joker.  It means "this is string data, but
+	 * the user may have really wanted it to be a number. If we have
+	 * to guess, like in a comparison, turn it into a number."
+	 * For example,    gawk -v a=42 ....
+	 * Here, `a' gets STRING|STRCUR|MAYBE_NUM and then when used where
+	 * a number is needed, it gets turned into a NUMBER and STRING
+	 * is cleared.
+	 *
+	 * WSTRCUR is for efficiency. If in a multibyte locale, and we
+	 * need to do something character based (substr, length, etc.)
+	 * we create the corresponding wide character string and store it,
+	 * and add WSTRCUR to the flags so that we don't have to do the
+	 * conversion more than once.
+	 *
+	 * The NUMINT flag may be used with a value of any type -- NUMBER,
+	 * STRING, or STRNUM. It indicates that the string representation
+	 * equals the result of sprintf("%ld", <numeric value>). So, for
+	 * example, NUMINT should NOT be set if it's a strnum or string value
+	 * where the string is " 1" or "01" or "+1" or "1.0" or "0.1E1". This
+	 * is a hint to indicate that an integer array optimization may be
+	 * used when this value appears as a subscript.
+	 *
+	 * We hope that the rest of the flags are self-explanatory. :-)
+	 */
 #		define	STRING	0x0002       /* assigned as string */
 #		define	STRCUR	0x0004       /* string value is current */
 #		define	NUMCUR	0x0008       /* numeric value is current */
@@ -1051,6 +1077,9 @@ extern afunc_t str_array_func[];
 extern afunc_t cint_array_func[];
 extern afunc_t int_array_func[];
 
+/* special node used to indicate success in array routines (not NULL) */
+extern NODE *success_node;
+
 extern BLOCK nextfree[];
 extern bool field0_valid;
 
@@ -1250,7 +1279,7 @@ DEREF(NODE *r)
 
 #define efree(p)	free(p)
 
-#define fatal		set_loc(__FILE__, __LINE__), r_fatal
+#define fatal		(*(set_loc(__FILE__, __LINE__), r_fatal))
 
 extern jmp_buf fatal_tag;
 extern bool fatal_tag_valid;
@@ -1420,9 +1449,9 @@ extern void close_extensions(void);
 extern void make_old_builtin(const char *, NODE *(*)(int), int);
 extern awk_bool_t make_builtin(const awk_ext_func_t *);
 extern NODE *get_argument(int);
-extern NODE *get_actual_argument(int, bool, bool);
-#define get_scalar_argument(i, opt)  get_actual_argument((i), (opt), false)
-#define get_array_argument(i, opt)   get_actual_argument((i), (opt), true)
+extern NODE *get_actual_argument(NODE *, int, bool);
+#define get_scalar_argument(n, i)  get_actual_argument((n), (i), false)
+#define get_array_argument(n, i)   get_actual_argument((n), (i), true)
 #endif
 /* field.c */
 extern void init_fields(void);
@@ -1483,6 +1512,8 @@ extern struct redirect *redirect(NODE *redir_exp, int redirtype, int *errflg);
 extern NODE *do_close(int nargs);
 extern int flush_io(void);
 extern int close_io(bool *stdio_problem);
+typedef enum { CLOSE_ALL, CLOSE_TO, CLOSE_FROM } two_way_close_type;
+extern int close_rp(struct redirect *rp, two_way_close_type how);
 extern int devopen_simple(const char *name, const char *mode, bool try_real_open);
 extern int devopen(const char *name, const char *mode);
 extern int srcopen(SRCFILE *s);
@@ -1633,6 +1664,11 @@ extern uintmax_t adjust_uint(uintmax_t n);
 #endif /* ! defined(VMS)) */
 #endif /* WEXITSTATUS */
 
+/* For z/OS, from Dave Pitts. EXIT_FAILURE is normally 8, make it 1. */
+#if defined(EXIT_FAILURE) && EXIT_FAILURE == 8
+# undef EXIT_FAILURE
+#endif
+
 /* EXIT_SUCCESS and EXIT_FAILURE normally come from <stdlib.h> */
 #ifndef EXIT_SUCCESS
 # define EXIT_SUCCESS 0
@@ -1645,16 +1681,6 @@ extern uintmax_t adjust_uint(uintmax_t n);
 # define EXIT_FATAL   2
 #endif
 
-/* For z/OS, from Dave Pitts. EXIT_FAILURE is normally 8, make it 1. */
-#ifdef ZOS_USS
-
-#ifdef EXIT_FAILURE
-#undef EXIT_FAILURE
-#endif
-
-#define EXIT_FAILURE 1
-#endif
-
 /* ------------------ Inline Functions ------------------ */
 
 /*
@@ -1745,6 +1771,8 @@ dupnode(NODE *n)
 }
 #endif
 
+/* force_string --- force a node to have a string value */
+
 static inline NODE *
 force_string(NODE *s)
 {
@@ -1760,6 +1788,8 @@ force_string(NODE *s)
 #define	force_number	str2number
 #else /* not GAWKDEBUG */
 
+/* unref --- decrease the reference count and/or free a node */
+
 static inline void
 unref(NODE *r)
 {
@@ -1767,6 +1797,8 @@ unref(NODE *r)
 		r_unref(r);
 }
 
+/* force_number --- force a  node to have a numeric value */
+
 static inline NODE *
 force_number(NODE *n)
 {
@@ -1775,6 +1807,8 @@ force_number(NODE *n)
 
 #endif /* GAWKDEBUG */
 
+/* emalloc_real --- malloc with error checking */
+
 static inline void *
 emalloc_real(size_t count, const char *where, const char *var, const char *file, int line)
 {
@@ -1791,6 +1825,8 @@ emalloc_real(size_t count, const char *where, const char *var, const char *file,
 	return ret;
 }
 
+/* erealloc_real --- realloc with error checking */
+
 static inline void *
 erealloc_real(void *ptr, size_t count, const char *where, const char *var, const char *file, int line)
 {
